@@ -2,17 +2,20 @@
  * SPDX-License-Identifier: MIT OR Apache-2.0
  */
 
-use super::{Action, CargoUI, DependencyNode, Diag, Feature};
-use cargo_metadata::{diagnostic::DiagnosticLevel, Metadata, Node, PackageId};
+use super::{Action, CargoUI, CratesCompletionData, DependencyData, DependencyNode, Diag, Feature};
+use anyhow::Context;
+use cargo_metadata::{
+    diagnostic::DiagnosticLevel, DependencyKind, Metadata, Node, PackageId, Version,
+};
+use futures::future::{Fuse, FusedFuture, FutureExt};
 use itertools::Itertools;
 use serde::Deserialize;
 use sixtyfps::{ComponentHandle, Model, ModelHandle, SharedString, VecModel};
 use std::cell::RefCell;
-use std::collections::{HashMap, HashSet};
-use std::future::Future;
+use std::collections::{HashMap, HashSet, VecDeque};
 use std::path::{Path, PathBuf};
-use std::pin::Pin;
 use std::rc::Rc;
+use std::str::FromStr;
 use tokio::io::{AsyncBufReadExt, BufReader};
 use tokio::sync::mpsc::{UnboundedReceiver, UnboundedSender};
 
@@ -33,6 +36,24 @@ pub enum CargoMessage {
     PackageSelected(SharedString),
     ShowOpenDialog,
     Cancel,
+    /// Remove the dependency `.1` from package `.0`
+    DependencyRemove {
+        parent_package: SharedString,
+        crate_name: SharedString,
+        dep_kind: DependencyKind,
+    },
+    /// Upgrade the dependency `.1` in package `.0`
+    DependencyUpgrade {
+        parent_package: SharedString,
+        crate_name: SharedString,
+        dep_kind: DependencyKind,
+    },
+    DependencyAdd {
+        crate_name: SharedString,
+        dep_kind: DependencyKind,
+    },
+    Install(InstallJob),
+    UpdateCompletion(SharedString),
 }
 
 pub struct CargoWorker {
@@ -49,6 +70,7 @@ impl CargoWorker {
                 tokio::runtime::Runtime::new()
                     .unwrap()
                     .block_on(cargo_worker_loop(r, handle_weak))
+                    .unwrap()
             }
         });
         Self {
@@ -66,84 +88,233 @@ impl CargoWorker {
 async fn cargo_worker_loop(
     mut r: UnboundedReceiver<CargoMessage>,
     handle: sixtyfps::Weak<CargoUI>,
-) {
+) -> tokio::io::Result<()> {
     let mut manifest: Manifest = default_manifest().into();
-    let metadata: RefCell<Option<Metadata>> = Default::default();
+    let mut metadata: Option<Metadata> = None;
+    let mut crates_index: Option<crates_index::Index> = None;
     let mut package = SharedString::default();
     let mut update_features = true;
-    let mut run_cargo_future: Option<Pin<Box<dyn Future<Output = tokio::io::Result<()>>>>> = Some(
-        Box::pin(read_metadata(manifest.clone(), handle.clone(), &metadata)),
+    let mut install_queue = VecDeque::new();
+    let mut currently_installing = SharedString::default();
+
+    let run_cargo_future = Fuse::terminated();
+    let read_metadata_future = read_metadata(manifest.clone(), handle.clone()).fuse();
+    let load_crate_index_future = load_crate_index().fuse();
+    let install_completion_future = Fuse::terminated();
+    let refresh_install_list_future = refresh_install_list(handle.clone()).fuse();
+    let process_install_future = Fuse::terminated();
+    futures::pin_mut!(
+        run_cargo_future,
+        read_metadata_future,
+        load_crate_index_future,
+        refresh_install_list_future,
+        process_install_future,
+        install_completion_future,
     );
     loop {
-        let m = if let Some(fut) = &mut run_cargo_future {
-            tokio::select! {
-                m = r.recv() => {
-                    m
+        let m = futures::select! {
+            res = run_cargo_future => {
+                res?;
+                continue;
+            }
+            res = read_metadata_future => {
+                metadata = res;
+                if let Some(metadata) = &metadata {
+                    apply_metadata(metadata, crates_index.as_ref(), update_features, &mut package, handle.clone());
+                    update_features = false;
                 }
-                res = fut => {
-                    res.unwrap();
-                    run_cargo_future = None;
-                    if let Some(metadata) = &*metadata.borrow() {
-                        apply_metadata(metadata, update_features, &mut package, handle.clone());
-                        update_features = false;
-                    }
-                    r.recv().await
+                continue;
+            }
+            index = load_crate_index_future => {
+                match index {
+                    Ok(idx) => crates_index = Some(idx),
+                    // TODO: ideally we should show that in the UI somehow
+                    Err(error) => eprintln!("Error while fetching crate index: {}", error),
+                };
+                if let Some(metadata) = &metadata {
+                    apply_metadata(metadata, crates_index.as_ref(), update_features, &mut package, handle.clone());
+                    update_features = false;
+                }
+                if refresh_install_list_future.is_terminated() {
+                    refresh_install_list_future.set(refresh_install_list(handle.clone()).fuse());
+                }
+                continue;
+            }
+            res = refresh_install_list_future =>  {
+                apply_install_list(res?, crates_index.as_ref(), &install_queue, &currently_installing, handle.clone());
+                continue;
+            }
+            res = process_install_future => {
+                res?;
+                refresh_install_list_future.set(refresh_install_list(handle.clone()).fuse());
+                if let Some(job) = install_queue.pop_front() {
+                    currently_installing = job.crate_name().clone();
+                    process_install_future.set(process_install(job, handle.clone()).fuse());
+                } else {
+                    currently_installing = Default::default();
+                }
+                continue;
+            }
+            _ = install_completion_future => { continue; }
+            m = r.recv().fuse() => {
+                match m {
+                    None => return Ok(()),
+                    Some(m) => m,
                 }
             }
-        } else {
-            r.recv().await
         };
 
         match m {
-            None => return,
-            Some(CargoMessage::Quit) => return,
-            Some(CargoMessage::Action {
+            CargoMessage::Quit => return Ok(()),
+            CargoMessage::Action {
                 action,
                 feature_settings,
-            }) => {
-                run_cargo_future = Some(Box::pin(run_cargo(
-                    action,
-                    feature_settings,
-                    manifest.clone(),
-                    handle.clone(),
-                )));
+            } => run_cargo_future
+                .set(run_cargo(action, feature_settings, manifest.clone(), handle.clone()).fuse()),
+            CargoMessage::Cancel => {
+                run_cargo_future.set(Fuse::terminated());
             }
-            Some(CargoMessage::Cancel) => {
-                run_cargo_future = None;
-            }
-            Some(CargoMessage::ReloadManifest(m)) => {
+            CargoMessage::ReloadManifest(m) => {
                 manifest = PathBuf::from(m.as_str()).into();
                 update_features = true;
-                run_cargo_future = Some(Box::pin(read_metadata(
-                    manifest.clone(),
-                    handle.clone(),
-                    &metadata,
-                )));
+                read_metadata_future.set(read_metadata(manifest.clone(), handle.clone()).fuse());
             }
-            Some(CargoMessage::ShowOpenDialog) => {
+            CargoMessage::ShowOpenDialog => {
                 manifest = show_open_dialog(manifest);
                 update_features = true;
-                run_cargo_future = Some(Box::pin(read_metadata(
-                    manifest.clone(),
-                    handle.clone(),
-                    &metadata,
-                )));
+                read_metadata_future.set(read_metadata(manifest.clone(), handle.clone()).fuse());
             }
-            Some(CargoMessage::PackageSelected(pkg)) => {
+            CargoMessage::PackageSelected(pkg) => {
                 package = pkg;
-                if let Some(metadata) = &*metadata.borrow() {
+                if let Some(metadata) = &metadata {
                     apply_metadata(
                         metadata,
+                        crates_index.as_ref(),
                         /*update_features*/ true,
                         &mut package,
                         handle.clone(),
                     );
                 }
             }
+            CargoMessage::DependencyRemove {
+                parent_package,
+                crate_name,
+                dep_kind,
+            } => {
+                let pkg = PackageId {
+                    repr: parent_package.into(),
+                };
+                if let Some(pkg) = metadata
+                    .as_ref()
+                    .and_then(|metadata| metadata.packages.iter().find(|p| p.id == pkg))
+                {
+                    match dependency_remove(
+                        pkg.manifest_path.as_ref(),
+                        crate_name.as_str(),
+                        dep_kind,
+                    ) {
+                        Ok(()) => read_metadata_future
+                            .set(read_metadata(manifest.clone(), handle.clone()).fuse()),
+                        Err(e) => {
+                            dbg!(e);
+                            handle.clone().upgrade_in_event_loop(|h| {
+                                h.set_status("Not yet supported".into());
+                            });
+                        }
+                    }
+                }
+            }
+            CargoMessage::DependencyAdd {
+                crate_name,
+                dep_kind,
+            } => {
+                if let Some((pkg, cr)) = metadata
+                    .as_ref()
+                    .and_then(|metadata| {
+                        let pkg = package.as_str();
+                        if pkg.is_empty() {
+                            Some(&metadata[metadata.workspace_members.first()?])
+                        } else {
+                            metadata.packages.iter().find(|p| p.name == pkg)
+                        }
+                    })
+                    .and_then(|p| Some((p, crates_index.as_ref()?.crate_(&crate_name)?)))
+                {
+                    match dependency_add(
+                        pkg.manifest_path.as_ref(),
+                        crate_name.as_str(),
+                        cr.highest_stable_version()
+                            .unwrap_or(cr.highest_version())
+                            .version(),
+                        dep_kind,
+                    ) {
+                        Ok(()) => read_metadata_future
+                            .set(read_metadata(manifest.clone(), handle.clone()).fuse()),
+                        Err(e) => {
+                            handle.clone().upgrade_in_event_loop(move |h| {
+                                h.set_status(format!("{}", e).into());
+                            });
+                        }
+                    }
+                }
+            }
+            CargoMessage::DependencyUpgrade {
+                parent_package,
+                crate_name,
+                dep_kind,
+            } => {
+                let pkg = PackageId {
+                    repr: parent_package.into(),
+                };
+                if let Some((pkg, cr)) = metadata
+                    .as_ref()
+                    .and_then(|metadata| metadata.packages.iter().find(|p| p.id == pkg))
+                    .and_then(|p| Some((p, crates_index.as_ref()?.crate_(&crate_name)?)))
+                {
+                    match dependency_upgrade_to_version(
+                        pkg.manifest_path.as_ref(),
+                        crate_name.as_str(),
+                        cr.highest_stable_version()
+                            .unwrap_or(cr.highest_version())
+                            .version(),
+                        dep_kind,
+                    ) {
+                        Ok(()) => read_metadata_future
+                            .set(read_metadata(manifest.clone(), handle.clone()).fuse()),
+                        Err(e) => {
+                            dbg!(e);
+                            handle.clone().upgrade_in_event_loop(|h| {
+                                h.set_status("Not yet supported".into());
+                            });
+                        }
+                    }
+                }
+            }
+            CargoMessage::Install(job) => {
+                if process_install_future.is_terminated() {
+                    currently_installing = job.crate_name().clone();
+                    process_install_future.set(process_install(job, handle.clone()).fuse());
+                } else {
+                    install_queue.push_back(job);
+                }
+            }
+            CargoMessage::UpdateCompletion(query) => {
+                if let Some(idx) = crates_index.as_ref() {
+                    install_completion_future.set(
+                        install_completion(idx.path().to_owned(), query, handle.clone()).fuse(),
+                    );
+                }
+            }
         }
     }
 }
 
+async fn load_crate_index() -> Result<crates_index::Index, String> {
+    let mut index = crates_index::Index::new_cargo_default().map_err(|x| x.to_string())?;
+    index.update().map_err(|x| x.to_string())?;
+    Ok(index)
+}
+
 async fn run_cargo(
     action: Action,
     features: FeatureSettings,
@@ -169,8 +340,7 @@ async fn run_cargo(
     }
     let _reset_is_building = ResetIsBuilding(handle.clone());
 
-    let cargo_path = std::env::var("CARGO").unwrap_or_else(|_| "cargo".into());
-    let mut cargo_command = tokio::process::Command::new(cargo_path);
+    let mut cargo_command = cargo_command();
     cargo_command.arg(action.command.as_str());
     cargo_command
         .arg("--manifest-path")
@@ -276,6 +446,11 @@ async fn run_cargo(
     Ok(())
 }
 
+pub fn cargo_command() -> tokio::process::Command {
+    let cargo_path = std::env::var("CARGO").unwrap_or_else(|_| "cargo".into());
+    tokio::process::Command::new(cargo_path)
+}
+
 fn cargo_message_to_diag(msg: cargo_metadata::Message) -> Option<Diag> {
     match msg {
         cargo_metadata::Message::CompilerMessage(msg) => {
@@ -319,11 +494,7 @@ fn default_manifest() -> PathBuf {
     .unwrap_or_default()
 }
 
-async fn read_metadata(
-    manifest: Manifest,
-    handle: sixtyfps::Weak<CargoUI>,
-    output: &RefCell<Option<Metadata>>,
-) -> tokio::io::Result<()> {
+async fn read_metadata(manifest: Manifest, handle: sixtyfps::Weak<CargoUI>) -> Option<Metadata> {
     let manifest_str = manifest
         .path_to_cargo_toml()
         .to_string_lossy()
@@ -337,7 +508,7 @@ async fn read_metadata(
 
     let mut cmd = cargo_metadata::MetadataCommand::new();
     cmd.manifest_path(manifest.path_to_cargo_toml());
-    *output.borrow_mut() = match cmd.exec() {
+    match cmd.exec() {
         Ok(metadata) => {
             handle.upgrade_in_event_loop(move |h| {
                 h.set_status("Cargo.toml loaded".into());
@@ -350,12 +521,12 @@ async fn read_metadata(
             });
             None
         }
-    };
-    Ok(())
+    }
 }
 
 fn apply_metadata(
     metadata: &Metadata,
+    crates_index: Option<&crates_index::Index>,
     mut update_features: bool,
     package: &mut SharedString,
     handle: sixtyfps::Weak<CargoUI>,
@@ -376,6 +547,8 @@ fn apply_metadata(
         update_features = true;
     };
 
+    let is_workspace = metadata.workspace_members.len() > 1;
+
     for p in metadata
         .packages
         .iter()
@@ -383,6 +556,8 @@ fn apply_metadata(
     {
         packages.push(p.name.as_str().into());
 
+        let is_selected = !is_workspace || package.is_empty() || package == p.name.as_str();
+
         if update_features {
             let default_features: HashSet<_> = p
                 .features
@@ -390,14 +565,14 @@ fn apply_metadata(
                 .map(|default_features| default_features.iter().cloned().collect())
                 .unwrap_or_default();
 
-            if package.is_empty() || package == p.name.as_str() {
+            if is_selected {
                 // Use get_or_insert_with_default() when https://github.com/rust-lang/rust/issues/82901 is stable
                 features.get_or_insert_with(|| Default::default()).extend(
                     p.features
                         .keys()
                         .filter(|name| *name != "default")
                         .map(|name| Feature {
-                            name: if package.is_empty() {
+                            name: if package.is_empty() && is_workspace {
                                 [p.name.as_str(), name.as_str()].join("/").into()
                             } else {
                                 name.into()
@@ -409,7 +584,7 @@ fn apply_metadata(
             }
         }
 
-        if !package.is_empty() && package != p.name.as_str() {
+        if !is_selected {
             continue;
         }
         for t in &p.targets {
@@ -424,17 +599,18 @@ fn apply_metadata(
     }
     let pkg = package.clone();
     handle.clone().upgrade_in_event_loop(move |h| {
+        h.global::<DependencyData>()
+            .set_package_selected(!is_workspace || !pkg.is_empty());
         h.set_current_package(pkg);
         // The model always has at least two entries, one for all and the first package,
         // so enable multi-package selection only if there is something else to select.
-        h.set_allow_package_selection(packages.len() > 2);
+        h.set_allow_package_selection(is_workspace);
         h.set_packages(ModelHandle::from(
             Rc::new(VecModel::from(packages)) as Rc<dyn Model<Data = SharedString>>
         ));
         h.set_extra_run(ModelHandle::from(
             Rc::new(VecModel::from(run_target)) as Rc<dyn Model<Data = SharedString>>
         ));
-        h.set_has_extra_tests(!test_target.is_empty());
         h.set_extra_test(ModelHandle::from(
             Rc::new(VecModel::from(test_target)) as Rc<dyn Model<Data = SharedString>>
         ));
@@ -459,9 +635,11 @@ fn apply_metadata(
             build_dep_tree(
                 m,
                 None,
+                &SharedString::default(),
                 &mut depgraph_tree,
                 &mut duplicates,
                 &metadata,
+                crates_index,
                 &map,
                 0,
             );
@@ -470,7 +648,8 @@ fn apply_metadata(
 
     handle.upgrade_in_event_loop(move |h| {
         let model = Rc::new(DepGraphModel::from(depgraph_tree));
-        h.set_deptree(ModelHandle::new(model))
+        h.global::<DependencyData>()
+            .set_model(ModelHandle::new(model))
     });
 }
 
@@ -500,6 +679,7 @@ fn show_open_dialog(manifest: Manifest) -> Manifest {
 #[cfg(any(target_os = "macos", target_os = "windows"))]
 fn show_open_dialog(manifest: Manifest) -> Manifest {
     let mut dialog = rfd::FileDialog::new();
+    dialog = dialog.set_title("Select a manifest");
 
     if let Some(directory) = manifest.directory() {
         dialog = dialog.set_directory(directory);
@@ -519,17 +699,23 @@ struct TreeNode {
 fn build_dep_tree(
     package_id: &PackageId,
     node_dep: Option<&cargo_metadata::NodeDep>,
+    parent_package: &SharedString,
     depgraph_tree: &mut Vec<TreeNode>,
     duplicates: &mut HashSet<PackageId>,
     metadata: &Metadata,
+    crates_index: Option<&crates_index::Index>,
     map: &HashMap<PackageId, &Node>,
     indentation: i32,
 ) {
     let package = &metadata[package_id];
-    let mut text = format!("{} {}", package.name, package.version).into();
-    if duplicates.contains(package_id) {
-        text += " (duplicated)";
-    }
+    let duplicated = duplicates.contains(package_id);
+    // We only consider indentation ==1 because `idx.crate_` is a bit too slow to do for every crate
+    let outdated = indentation == 1
+        && crates_index
+            .and_then(|idx| idx.crate_(&package.name))
+            .and_then(|c| c.highest_stable_version().cloned())
+            .and_then(|v| Version::from_str(v.version()).ok())
+            .map_or(false, |latest| latest > package.version);
     let dep_kind = node_dep
         .filter(|n| {
             !n.dep_kinds
@@ -549,8 +735,12 @@ fn build_dep_tree(
             has_children: false,
             indentation,
             open: indentation != 1,
-            text,
+            version: package.version.to_string().into(),
+            crate_name: package.name.as_str().into(),
+            outdated,
+            duplicated,
             dep_kind,
+            parent_package: parent_package.clone(),
         }
         .into(),
         children: Default::default(),
@@ -563,9 +753,11 @@ fn build_dep_tree(
             build_dep_tree(
                 &d.pkg,
                 Some(&d),
+                &package_id.repr.as_str().into(),
                 &mut node.children,
                 duplicates,
                 metadata,
+                crates_index,
                 map,
                 indentation + 1,
             );
@@ -641,8 +833,8 @@ impl sixtyfps::Model for DepGraphModel {
         node.node.borrow().clone()
     }
 
-    fn attach_peer(&self, peer: sixtyfps::ModelPeer) {
-        self.notify.attach(peer);
+    fn model_tracker(&self) -> &dyn sixtyfps::ModelTracker {
+        &self.notify
     }
 
     fn set_row_data(&self, row: usize, data: Self::Data) {
@@ -706,3 +898,109 @@ impl FeatureSettings {
         }
     }
 }
+
+fn to_table_name(dep_kind: DependencyKind) -> &'static str {
+    match dep_kind {
+        DependencyKind::Development => "dev-dependencies",
+        DependencyKind::Build => "build-dependencies",
+        _ => "dependencies",
+    }
+}
+
+fn dependency_remove(pkg: &Path, dependency: &str, dep_kind: DependencyKind) -> anyhow::Result<()> {
+    let manifest_contents = std::fs::read_to_string(pkg)
+        .with_context(|| format!("Failed to load '{}'", pkg.display()))?;
+    let mut document: toml_edit::Document = manifest_contents.parse()?;
+    let table_name = to_table_name(dep_kind);
+    let dependencies = &mut document[table_name];
+    let removed = !std::mem::take(&mut dependencies[dependency]).is_none();
+    if !removed {
+        anyhow::bail!("'{}' was not in [{}]", dependency, dep_kind);
+    }
+    std::fs::write(pkg, document.to_string().as_bytes())
+        .with_context(|| format!("Failed to write '{}'", pkg.display()))
+}
+
+fn dependency_upgrade_to_version(
+    pkg: &Path,
+    dependency: &str,
+    version: &str,
+    dep_kind: DependencyKind,
+) -> anyhow::Result<()> {
+    let manifest_contents = std::fs::read_to_string(pkg)
+        .with_context(|| format!("Failed to load '{}'", pkg.display()))?;
+    let mut document: toml_edit::Document = manifest_contents.parse()?;
+    let table_name = to_table_name(dep_kind);
+    let dep = &mut document[table_name][dependency];
+    if dep.is_none() {
+        anyhow::bail!("'{}' was not in [{}]", dependency, dep_kind);
+    }
+    if dep.is_str() {
+        *dep = toml_edit::Item::Value(version.into());
+    } else if dep.is_table_like() {
+        dep["version"] = toml_edit::Item::Value(version.into());
+    } else {
+        anyhow::bail!("Could not understand the manifest");
+    }
+    std::fs::write(pkg, document.to_string().as_bytes())
+        .with_context(|| format!("Failed to write '{}'", pkg.display()))
+}
+
+fn dependency_add(
+    pkg: &Path,
+    dependency: &str,
+    version: &str,
+    dep_kind: DependencyKind,
+) -> anyhow::Result<()> {
+    let manifest_contents = std::fs::read_to_string(pkg)
+        .with_context(|| format!("Failed to load '{}'", pkg.display()))?;
+    let mut document: toml_edit::Document = manifest_contents.parse()?;
+    let table_name = to_table_name(dep_kind);
+    let tbl = &mut document[table_name].or_insert(toml_edit::table());
+    if !tbl.is_table_like() {
+        anyhow::bail!("[{}] not a table", table_name);
+    }
+    let dep = &mut tbl[dependency];
+    if !dep.is_none() {
+        anyhow::bail!("{} is already a dependency", dependency);
+    }
+    *dep = toml_edit::Item::Value(version.into());
+
+    std::fs::write(pkg, document.to_string().as_bytes())
+        .with_context(|| format!("Failed to write '{}'", pkg.display()))
+}
+
+use crate::install::*;
+
+async fn install_completion(
+    mut idx_path: PathBuf,
+    query: SharedString,
+    handle: sixtyfps::Weak<CargoUI>,
+) {
+    let mut result = Vec::<SharedString>::new();
+    if query.len() > 3 && query.is_ascii() {
+        // `crates_index` does not allow to make search in a reasonable time, so I had to implement that myself
+        // This only handle crates that have 4 or more characters
+        idx_path.push(&query[0..2]);
+        idx_path.push(&query[2..4]);
+        if let Ok(mut rd) = tokio::fs::read_dir(idx_path).await {
+            while let Ok(Some(entry)) = rd.next_entry().await {
+                if let Some(name) = entry.file_name().to_str() {
+                    if name.starts_with(query.as_str()) {
+                        result.push(name.into());
+                        if result.len() > 50 {
+                            // no need to put too many crate in the search result
+                            break;
+                        }
+                    }
+                }
+            }
+        }
+    }
+    handle.upgrade_in_event_loop(move |ui| {
+        ui.global::<CratesCompletionData>()
+            .set_completions(ModelHandle::from(
+                Rc::new(VecModel::from(result)) as Rc<dyn Model<Data = SharedString>>
+            ));
+    });
+}
